#include "windows.inc"
//#include "commctrl.inc"
#autodefine "off"
#define DEFAULT_TEXT_LIMIT 65536
#define COLOREDIT_MIN_CHARS_IN_LINE 256 // for horizontal scrollbar minimum range
#define COLOREDIT_HAS_CONTEXT_MENU


//#define MOVE_CARET_WITH_PAGE_UPDOWN // enable if PageUp or PageDown key should move the caret
//#define USE_COLOREDIT_EXTRAINFO     // enable if you want to know what is the longest line (struct COLOREDIT_EXTRAINFO in CColorEdit::m_extrainfo)


#ifdef COLOREDIT_HAS_CONTEXT_MENU
	#define IDM_COLOREDIT_CUT       1000
	#define IDM_COLOREDIT_COPY      1001
	#define IDM_COLOREDIT_PASTE     1002
	#define IDM_COLOREDIT_SELECTALL 1003
	#define IDM_COLOREDIT_DELETE    1004
	#define IDM_COLOREDIT_FIND      1005
	#define IDM_COLOREDIT_FINDNEXT  1006
	#define IDM_COLOREDIT_REPLACE   1007

	#define IDM_COLOREDIT_EDIT_RELEASE_NOTES 5000;
#endif

//---------------------------------------------------------------
// Keywords classes:
// struct KEYWORD_LIST   - pointers to words in editbox text
// class CKeywordsGroup  - how to color a word
// class CKeywordsRoot   - add and find: keywords and colors
//---------------------------------------------------------------

struct KEYWORD_LIST
{
	KEYWORD_LIST *next;
	int     m_len;      // searched string size = m_len? yes: compare string with m_str
	dstring m_str[1];
}

//---------------------------------------------------------------
// this class stores groups of keywords, where a group can have different colors
// eg. group virtually named "keywords" has blue keywords like declare, virtual, sub ...
//     group virtually named "variables" has red keywords like int, byte ...
// Pointer to next group is stored in .next
// Pointer to keywords (with same colors) is stored in .m_pKeywords
/*
sub CKeywordsGroup
*/
class CKeywordsGroup
{
	// colors and font
	COLORREF        m_TextColor;
	COLORREF        m_TextBkColor;
	HFONT           m_hFont;       // not owned, default null

	// keywords and other colorized groups
	KEYWORD_LIST   *m_pKeywords;
	CKeywordsGroup *next;          // for root class: this points o first group (with valid m_extrainfo)

	// speedup searching: compare two dwords is faster than compare all strings
	word wMinKeywordLen, wMaxKeywordLen;

	declare CKeywordsGroup();
	declare _CKeywordsGroup();
	declare AddKeyword(string *keyword,opt int cch=-1),BOOL;
	declare AddKeywords(string *keywords),BOOL; // separate strings by non-alphabetic character: "do until declare"

}

// this class manages groups ( linked list ofclasses CKeywordsGroup)
/*
sub CKeywordsRoot
*/
class CKeywordsRoot
{
	declare CKeywordsRoot();
	declare _CKeywordsRoot();

	declare AddGroup(),CKeywordsGroup*;
	declare Find(string *keyword,opt int cch=-1),CKeywordsGroup *;

	CKeywordsGroup *m_groups;
}



//---------------------------------------------------------------
//
//---------------------------------------------------------------


// this class "divides" text into smaller tokens, where 
// every token stores information where begins n-th line, what color
// should be used to draw a word(s)
// To loop through lines - use .m_nextline pointer
// To loop within line through different colorized words (tokens) - use .m_nexttoken pointer
// Every line contains:
//  - pointer to beginning of text (m_lpText)
//  - size of the whole line (m_TextLen) including new line characters (if present)
//  - optional pointer to line tokens. If line begins with \n - no tokens are created.
// Every token points to beginning of text (.m_lpText) and token size (m_TextLen; in characters).
// To drawin a line we need to loop through all tokens, change text color and draw m_TextLen characters from m_lpText address.
// This technique does not need to duplicate every string token, it stores only address of text and text size.
// Example for text line: "sub    MySub(int x),int\n"
// line:
//  - m_lpText = "sub    MySub(int x),int\n" // strings are not duplicated
//  - m_TextLen = 25
// tokens:
//  - sub     // if "sub" was registered as keyword - color is changed; m_TextLen = 3
//  - 4 spaces// color back to normal text;    m_TextLen = 4
//  - MySub   // if registered as keyword ...; m_TextLen = 5
//  - {       // color back to normal text;    m_TextLen = 1
//  - int     // if registered as keyword ...; m_TextLen = 3
//  - 1 space // color back to normal text;    m_TextLen = 1
//  - x       // if registered as keyword ...; m_TextLen = 1
//  - ),      // color back to normal text;    m_TextLen = 2
//  - int     // if registered as keyword ...; m_TextLen = 3
// new line found!
// - m_nextline = new(CLineToken)
// - m_nextline.m_lpText = &"\n"
// - m_nextline.m_TextLen = number of characters from (&"\n" + 2) to (next "\n" or end of text)
/*
sub CLineToken
*/
class CLineToken
{
	declare  CLineToken(); // ZeroMemory(&m_nexline, 5<<2);
	declare _CLineToken(); // delete current line tokens (token.next) and all "child" lines 

	declare  AddLine(),CLineToken *;     // appends new line to end of list
	declare  AddToken(),CLineToken *;    // appends new token to current line
	declare  GetNextToken(),CLineToken*;
	declare  GetNextLine(),CLineToken*;

// 	CLineToken *m_prevline; ?? would be great
	CLineToken *m_nextline;    // prev/next line (CLineToken*)
	CLineToken *m_nexttoken;  //next token from this line
	string  *m_lpText;
	int      m_TextLen;
	CKeywordsGroup *m_pgroup;                        // todo: replace colors by this pointer
}


#typedef TOKENTYPE int;
const TOKENTYPE_NONE              = 0;
const TOKENTYPE_TEXT              = 1;
const TOKENTYPE_STRING            = 2;
const TOKENTYPE_COMMENT           = 3;
const TOKENTYPE_MULTILINE_COMMENT = 4;
const TOKENTYPE_DIGIT             = 5;
const TOKENTYPE_PREPROC           = 6;
const TOKENTYPE_NEWLINE           = 7;
const TOKENTYPE_KEYWORD           = 8;
const TOKENTYPE_LAST              = 8;


#typedef COLOREDITFLAG int;
const COLOREDITFLAG_LBUTTONDOWN = 0x00000001;


#ifdef USE_COLOREDIT_EXTRAINFO
struct COLOREDIT_EXTRAINFO
{
	int LongestLineNumber;    // CColorEdit::CalculateMaxLineWidth (zero based)
	int LongestLineCharacters;// CColorEdit::CalculateMaxLineWidth
	int NumberOfTextTokens;   // CColorEdit::ParseText
	int TextTokensMemorySize; // CColorEdit::ParseText
	int TextTokensDeleted;    // CColorEdit::ParseText
	int TextTokensMerged;     // CColorEdit::ParseText
}
#endif

/*
sub CCaret
*/
class CCaret
{
	declare CCaret();
	declare _CCaret();
	declare Create(HWND hwnd,int width,int height),BOOL;
	declare Destroy();
	declare Show(opt BOOL force=false);
	declare Hide();

	declare SetPos(int x,int y),BOOL;
	declare GetPos(),POINT *;
	declare Move(),BOOL; // set pos to current coords
	declare MoveX(int x);

	declare SetLine(int line);
	declare SetCharacter(int character);

	POINT m_pos;
	int   m_line;
	int   m_char;
	BOOL  m_created;
	int   m_hidecount;
	HWND  m_hwnd;
	int   m_xoffset;   // left margin for the caret. You can hange it.
	BOOL  m_hidden;    // if (m_pos.x < m_xoffset) hide 
}

class CSelection
{
	declare CSelection(); // ZeroMemory(&pLine,4<<2);

	CLineToken *pLine;
	CLineToken *pToken;
	int         LineIndex;
	int         CharIndex;
}

//-----------------------------------------------------------
// main class
// implemented in: CColorEdit.src
//
// todo: cut, copy, paste, replace, delete ...
// you can use (with no risk) only virtual methods (but standard)
/*
sub CColorEdit
*/
class CColorEdit : CWindow
{
	#ifdef USE_COLOREDIT_EXTRAINFO
	COLOREDIT_EXTRAINFO m_extrainfo;
	#endif

	//////////////////////////////////////////
	// standard (private methods)
	declare         CColorEdit();
	declare virtual OnCreate();
	declare virtual OnDestroy();
	declare virtual WndProc(unsigned int message, unsigned int wparam, unsigned int lparam),int;
	declare         OnEditMessage(unsigned int message, unsigned int wparam, unsigned int lparam),int;
	declare virtual OnSize(int nType,int cx,int cy),int;
	declare virtual OnMouseMove(int x,int y,int flags),int;
	declare virtual OnPaint(),int;
	declare virtual OnHScroll(int nCommand,int nPos,int nID),int;
	declare virtual OnVScroll(int nCommand,int nPos,int nID),int;
	declare virtual OnKeyDown(unsigned int nChar, INT nRepCnt, unsigned int nFlags),int;
	declare virtual OnLeftKeyDown();
	declare virtual OnRightKeyDown();
	declare virtual OnUpKeyDown();
	declare virtual OnDownKeyDown();
	declare virtual OnLButtonDown(int x,int y, int flags),int;
	declare virtual OnLButtonUp(int x,int y,int flags),int;
	declare virtual OnRButtonDown(int x,int y,int flags),int;
	declare virtual OnRButtonUp(int x,int y,int flags),int;
	#ifdef COLOREDIT_HAS_CONTEXT_MENU
	declare virtual OnMenuInit(unsigned int hMenu),int;
	declare virtual OnMenuPick(int nID),int;
	CMenu m_menu;
	#endif
	// other
	declare virtual OnFontChanged();      // call after SetFont to redraw and resize the caret
	declare virtual OnLineNumberChanged();// this method resizes linenumber box. Derieve it to prevent auto resizing.
	HCURSOR m_hCursorIbeam;
	HCURSOR m_hCursorCurrent;
	HCURSOR m_hCursorArrow;
	COLOREDITFLAG m_flags;

	//////////////////////////////////////////
	// drawing text

	declare virtual ForceRedraw();
	declare         OnBeforeDraw();            // prepare HDC and HFONT, recalculate line spacing
	declare         OnAfterDraw();             // cleanup HDC and HFONT
	declare         SetDrawRect(RECT *rcDraw); // ClientRect - margins
	declare         DrawAllLines();
	declare         DrawLines(int LineIndex, int count, RECT *rcDraw);
	declare         DrawLine(CLineToken *pLine, RECT *rcDraw);
	declare         GetLine(int lineindex),CLineToken *;                           // todo: make if faster
	declare         DrawLineNumber(int line, RECT *rcDraw);
	declare         DrawExchangeColors(CLineToken *line,opt BOOL back=false);
	declare virtual GetTextWidth(string *lpText,int textlen),int;
	HDC     m_dc;       // window HDC
	HDC     m_memDC;    // memory DC for text drawing; usefull for horizontal scrolling
	HBITMAP m_membmp;   // memory bitmap for lines
	HFONT   m_hFontCache; // equal to CWindow::m_hFont. If m_hFont hanges - OnFontChange is executed

	//////////////////////////////////////////
	// caret

	CCaret caret;
	declare         MoveCaretToCursor(int x,int y);
	declare         ScrollToCaretPos();
	declare         CursorPosToCharIndex(CLineToken *pLine, int xPixel/*client.cursor.x*/),int;

//	int  m_CaretPixelIndex;// caret pixel index for current line. OnLButtonDown()

	//////////////////////////////////////////
	// selection

	CSelection m_SelStart; // only those two structures!
	CSelection m_SelEnd;

	declare         SetSel(int start,int end);
	declare         RemoveSel();

	//////////////////////////////////////////
	// scrolling and scrollbars
	declare         UpdateScrollsInfo();
	declare virtual ScrollToLine(int line);
	int  m_hScrollPos;
	int  m_hScrollMax; // added now!
	int  m_MaxCharactersInLine;

	//////////////////////////////////////////
	// text handling

	declare virtual SetText(string *lpText,opt int size=-1),BOOL;
	declare virtual AppendText(string *lpText,opt int size=-1),BOOL;
	declare         CalculateMaxLineWidth();  // update m_MaxLineWidth
	declare         DeleteAllLines(); // frees m_lpstring and m_lines (auto)
	int  m_MaxCharWidth;

	//////////////////////////////////////////
	// parsing
	declare         ParseText(),BOOL;
	declare virtual ParseTextEx(CLineToken *pCurrentLine, byte *lpText);
	declare virtual ParserParseToken(byte *first byref, byte *last, TOKENTYPE type byref); // 'last' points to first character after text!
	declare virtual ParserGetStringTokenLength(byte *lpstr,byte *maxaddr),int;
	declare virtual ParserSetTokenColor(CLineToken *token, TOKENTYPE type);

	CLineToken m_lines; // points only to first line of text
	int  m_LinesCount;
	int  m_TopLineIndex; // 0 on startup
	int  m_LinesPerPage; // 0 on startup; updated in OnSize()
	int  m_MaxLineWidth;
	int  m_MaxLineChars;
	int  m_LineSpacing;  // adjusted automatically in OnBeforeDraw
	RECT m_rcSize;       // 0,0, {OnSize}

	//////////////////////////////////////////
	// keywords
	declare virtual GetKeywords(),CKeywordsRoot *;
	CKeywordsRoot m_keywords;

	// background color
	declare virtual SetBkColor(COLORREF color);  // cleanup(-1)
	COLORREF m_bkcolor;
	HBRUSH   m_hbrBk;

	//////////////////////////////////////////
	// color groups

	// normal text colors
	declare virtual SetTextBkColor(COLORREF color,opt BOOL redraw=true);
	declare virtual SetTextColor(COLORREF color,opt BOOL redraw=true);
	CKeywordsGroup m_GroupText;

	// string text color - "blah" or 'a'
	declare virtual SetStringColor(COLORREF color,opt BOOL redraw=true);
	CKeywordsGroup m_GroupString;

	// comment text color - /*blah*/ or //blah
	declare virtual SetCommentColor(COLORREF color,opt BOOL redraw=true);
	CKeywordsGroup m_GroupComment;

	// numerical text color - 0  or 0x67  or  1e6
	declare virtual SetNumColor(COLORREF color,opt BOOL redraw=true);
	CKeywordsGroup m_GroupNumeric;

	// preprocessor text color - #define  #if
	declare virtual SetPreprocColor(COLORREF color,opt BOOL redraw=true);
	CKeywordsGroup m_GroupPreproc;

	//////////////////////////////////////////
	// standalone colors

	// selected text color
	declare virtual SetSelTextColor(COLORREF color,opt BOOL redraw=true);
	COLORREF m_seltxtcolor;

	// selectedtext background color
	declare virtual SetSelTextBkColor(COLORREF color,opt BOOL redraw=true);
	COLORREF m_seltxtbkcolor;

	// global keyword text color - 
///	declare SetKeywordColor(COLORREF color,opt BOOL redraw=true);
///	COLORREF m_keywordcolor;

	// tabstop
	declare virtual SetTabSize(int tabsize,opt BOOL redraw=true);
	int  m_TabSize;      // 4  default
	int  m_TabSizeInPixels;

	// tokens
	byte m_byte_preproc; // default '#'
}




///////////////////////////
// messages:

//*WM_SETFONT    - incomplete
// WM_MOUSEWHEEL - scrolls the edit control
// WM_SETFOCUS   - displays the caret
// WM_KILLFOCUS  - hides the caret
// WM_SETTEXT    - replaces text

// EM_GETSEL     - retrieves the starting and ending character positions of the current selection
//*EM_SETSEL     - selects a range of characters




